"""A frozen immutable list module.

This module provides an immutable list type `Block` and  a set of
useful methods and functions for working with the list.

Named "Block" to avoid conflicts with the builtin Python List type.

A Block is actually backed by a Python tuple. Tuples in Python are
immutable and gives us a high performant implementation of immutable
lists.

Example:
    >>> xs = block.of_list([1, 2, 3, 4, 5])
    >>> ys = block.empty.cons(1).cons(2).cons(3).cons(4).cons(5)
    >>> zs = pipe(
...     xs,
...     block.filter(lambda x: x<10)
... )
"""
from __future__ import annotations

import builtins
import functools
import itertools
from collections.abc import Collection
from typing import (
    Any,
    Callable,
    Iterable,
    Iterator,
    List,
    Literal,
    Tuple,
    TypeVar,
    cast,
    overload,
)

from expression.core import (
    Nothing,
    Option,
    PipeMixin,
    Some,
    SupportsLessThan,
    SupportsSum,
    curry_flip,
    pipe,
)
from expression.core.typing import GenericValidator, ModelField, SupportsValidation

from . import seq

_TSource = TypeVar("_TSource")
_TSourceSortable = TypeVar("_TSourceSortable", bound=SupportsLessThan)
_TSourceSum = TypeVar("_TSourceSum", bound=SupportsSum)
_TResult = TypeVar("_TResult")
_TState = TypeVar("_TState")

_T1 = TypeVar("_T1")
_T2 = TypeVar("_T2")
_T3 = TypeVar("_T3")
_T4 = TypeVar("_T4")


def _validate(value: Any, field: ModelField) -> Block[Any]:
    if isinstance(value, Block):
        return cast(Block[Any], value)

    if not isinstance(value, List):
        raise ValueError("not a list")

    value_ = cast(List[Any], value)

    if field.sub_fields:
        sub_field = field.sub_fields[0]

        value__: List[Any] = []
        for item in value_:
            val, error = sub_field.validate(item, {}, loc="Block")
            if error:
                raise ValueError(str(error))
            value__.append(val)
        value_ = value__

    return Block(value_)


class Block(
    Collection[_TSource],  # Sequence breaks pydantic
    PipeMixin,
    SupportsValidation["Block[_TSource]"],
):
    """Immutable list type.

    Is faster than `List` for prepending, but slower for
    appending.

    Count: 200K::

        | Operation | Block      | List   |
        |-----------|------------|--------|
        | Append    | 3.29 s     | 0.02 s |
        | Prepend   | 0.05 s     | 7.02 s |

    Example:
        >>> xs = Cons(5, Cons(4, Cons(3, Cons(2, Cons(1, Nil)))))
        >>> ys = empty.cons(1).cons(2).cons(3).cons(4).cons(5)
    """

    __match_args__ = ("_value",)

    __validators__: List[GenericValidator[Block[_TSource]]] = [_validate]

    def __init__(self, value: Iterable[_TSource] = ()) -> None:
        # Use composition instead of inheritance since generic tuples
        # are not suppored by mypy.
        self._value: Tuple[_TSource, ...] = tuple(value) if value else tuple()

    def append(self, other: Block[_TSource]) -> Block[_TSource]:
        """Append other block to end of the block."""

        return Block(self._value + other._value)

    def choose(
        self, chooser: Callable[[_TSource], Option[_TResult]]
    ) -> Block[_TResult]:
        """Choose items from the list.

        Applies the given function to each element of the list. Returns
        the list comprised of the results x for each element where the
        function returns `Some(x)`.

        Args:
            chooser: The function to generate options from the elements.

        Returns:
            The list comprising the values selected from the chooser
            function.
        """

        def mapper(x: _TSource) -> Block[_TResult]:
            return Block(chooser(x).to_seq())

        return self.collect(mapper)

    def collect(
        self, mapping: Callable[[_TSource], Block[_TResult]]
    ) -> Block[_TResult]:
        mapped = builtins.map(mapping, self._value)
        xs = (y for x in mapped for y in x)
        return Block(xs)

    def cons(self, element: _TSource) -> Block[_TSource]:
        """Add element to front of list."""

        return Block((element,) + self._value)  # NOTE: Faster than (element, *self)

    @staticmethod
    def empty() -> Block[Any]:
        """Returns empty list."""

        return Block()

    def filter(self, predicate: Callable[[_TSource], bool]) -> Block[_TSource]:
        """Filter list.

        Returns a new collection containing only the elements of the
        collection for which the given predicate returns `True`.

        Args:
            predicate: The function to test the input elements.

        Returns:
            A list containing only the elements that satisfy the
            predicate.
        """
        return Block(builtins.filter(predicate, self._value))

    def fold(
        self, folder: Callable[[_TState, _TSource], _TState], state: _TState
    ) -> _TState:
        """Applies a function to each element of the collection,
        threading an accumulator argument through the computation. Take
        the second argument, and apply the function to it and the first
        element of the list. Then feed this result into the function
        along with the second element and so on. Return the final
        result. If the input function is f and the elements are i0...iN
        then computes f (... (f s i0) i1 ...) iN.

        Args:
            folder: The function to update the state given the input
                elements.

            state: The initial state.

        Returns:
            Partially applied fold function that takes the source list
            and returns the final state value.
        """
        return functools.reduce(folder, self, state)

    def forall(self, predicate: Callable[[_TSource], bool]) -> bool:
        """Tests if all elements of the collection satisfy the given
        predicate.

        Args:
            predicate: The function to test the input elements.

        Returns:
            True if all of the elements satisfy the predicate.
        """
        return all(predicate(x) for x in self)

    def head(self) -> _TSource:
        """Returns the first element of the list.

        Args:
            source: The input list.

        Returns:
            The first element of the list.

        Raises:
            ValueError: Thrown when the list is empty.
        """

        head, *_ = self
        return head

    def indexed(self, start: int = 0) -> Block[Tuple[int, _TSource]]:
        """Returns a new list whose elements are the corresponding
        elements of the input list paired with the index (from `start`)
        of each element.

        Args:
            start: Optional index to start from. Defaults to 0.

        Returns:
            The list of indexed elements.
        """
        return of_seq(enumerate(self))

    def item(self, index: int) -> _TSource:
        """Indexes into the list. The first element has index 0.

        Args:
            index: The index to retrieve.

        Returns:
            The value at the given index.
        """
        return self[index]

    def is_empty(self) -> bool:
        """Return `True` if list is empty."""

        return not bool(self)

    def map(self, mapping: Callable[[_TSource], _TResult]) -> Block[_TResult]:
        """Map list.

        Builds a new collection whose elements are the results of
        applying the given function to each of the elements of the
        collection.

        Args:
            mapping: The function to transform elements from the input
                list.

        Returns:
            The list of transformed elements.
        """
        return Block((*builtins.map(mapping, self),))

    @overload
    def starmap(
        self: Block[Tuple[_T1, _T2]], mapping: Callable[[_T1, _T2], _TResult]
    ) -> Block[_TResult]:
        ...

    @overload
    def starmap(
        self: Block[Tuple[_T1, _T2, _T3]],
        mapping: Callable[[_T1, _T2, _T3], _TResult],
    ) -> Block[_TResult]:
        ...

    @overload
    def starmap(
        self: Block[Tuple[_T1, _T2, _T3, _T4]],
        mapping: Callable[[_T1, _T2, _T3, _T4], _TResult],
    ) -> Block[_TResult]:
        ...

    def starmap(
        self: Block[Tuple[Any, ...]], mapping: Callable[..., Any]
    ) -> Block[Any]:
        """Starmap source sequence.

        Unpack arguments grouped as tuple elements. Builds a new collection
        whose elements are the results of applying the given function to the
        unpacked arguments to each of the elements of the collection.

        Args:
            mapping: A function to transform items from the input sequence.

        Returns:
            Partially applied map function.
        """
        return Block(starmap(mapping)(self))

    def sum(self: Block[_TSourceSum | Literal[0]]) -> _TSourceSum | Literal[0]:
        return builtins.sum(self._value)

    def sum_by(
        self: Block[_TSourceSum], projection: Callable[[_TSourceSum], _TResult]
    ) -> _TResult:
        return pipe(self, sum_by(projection))

    def mapi(self, mapping: Callable[[int, _TSource], _TResult]) -> Block[_TResult]:
        """Map list with index.

        Builds a new collection whose elements are the results of
        applying the given function to each of the elements of the
        collection. The integer index passed to the function indicates
        the index (from 0) of element being transformed.

        Args:
            mapping: The function to transform elements and their
                indices.

        Returns:
            The list of transformed elements.
        """
        return Block((*itertools.starmap(mapping, enumerate(self)),))

    @staticmethod
    def of(*args: _TSource) -> Block[_TSource]:
        """Create list from a number of arguments."""
        return Block((*args,))

    @staticmethod
    def of_seq(xs: Iterable[_TSource]) -> Block[_TSource]:
        """Create list from iterable sequence."""
        return Block((*xs,))

    @staticmethod
    def of_option(option: Option[_TSource]) -> Block[_TSource]:
        return of_option(option)

    def partition(
        self, predicate: Callable[[_TSource], bool]
    ) -> Tuple[Block[_TSource], Block[_TSource]]:
        """Partition block.

        Splits the collection into two collections, containing the
        elements for which the given predicate returns True and False
        respectively. Element order is preserved in both of the created
        lists.

        Args:
            predicate: The function to test the input elements.

        Returns:
            A list containing the elements for which the predicate
            evaluated to true and a list containing the elements for
            which the predicate evaluated to false.
        """
        list1: List[_TSource] = []
        list2: List[_TSource] = []

        for item in self._value:
            if predicate(item):
                list1.append(item)
            else:
                list2.append(item)
        return (Block(list1), Block(list2))

    @overload
    @staticmethod
    def range(stop: int) -> Block[int]:
        ...

    @overload
    @staticmethod
    def range(start: int, stop: int) -> Block[int]:
        ...

    @overload
    @staticmethod
    def range(start: int, stop: int, step: int) -> Block[int]:
        ...

    @staticmethod
    def range(*args: int, **kw: int) -> Block[int]:
        return range(*args, **kw)

    def reduce(
        self,
        reduction: Callable[[_TSource, _TSource], _TSource],
    ) -> _TSource:
        """Apply a function to each element of the collection, threading an
        accumulator argument through the computation. Apply the function to
        the first two elements of the list. Then feed this result into the
        function along with the third element and so on. Return the final
        result. If the input function is f and the elements are i0...iN then
        computes f (... (f i0 i1) i2 ...) iN.

        Args:
            reduction: The function to reduce two list elements to a
            single element.

        Returns:
            Returns the final state value.
        """
        return reduce(reduction)(self)

    @staticmethod
    def singleton(item: _TSource) -> Block[_TSource]:
        return singleton(item)

    def skip(self, count: int) -> Block[_TSource]:
        """Returns the list after removing the first N elements.

        Args:
            count: The number of elements to skip.

        Returns:
            The list after removing the first N elements.
        """
        return Block(self._value[count:])

    def skip_last(self, count: int) -> Block[_TSource]:
        return Block(self._value[:-count])

    def tail(self) -> Block[_TSource]:
        """Return tail of List."""

        _, *tail = self._value
        return Block(tail)

    def sort(
        self: Block[_TSourceSortable], reverse: bool = False
    ) -> Block[_TSourceSortable]:
        """Sort list directly.

        Returns a new sorted collection.

        Args:
            reverse: Sort in reversed order.

        Returns:
            A sorted list.
        """
        return Block(builtins.sorted(self._value, reverse=reverse))

    def sort_with(
        self, func: Callable[[_TSource], Any], reverse: bool = False
    ) -> Block[_TSource]:
        """Sort list with supplied function.

        Returns a new sorted collection.

        Args:
            func: The function to extract a comparison key from each element in list.
            reverse: Sort in reversed order.

        Returns:
            A sorted list.
        """
        return Block(builtins.sorted(self._value, key=func, reverse=reverse))

    def take(self, count: int) -> Block[_TSource]:
        """Returns the first N elements of the list.

        Args:
            count: The number of items to take.

        Returns:
            The result list.
        """
        return Block(self._value[:count])

    def take_last(self, count: int) -> Block[_TSource]:
        """Returns a specified number of contiguous elements from the
        end of the list.

        Args:
            count: The number of items to take.

        Returns:
            The result list.
        """
        return Block(self._value[-count:])

    def dict(self) -> List[_TSource]:
        """Returns a json serializable representation of the list."""

        def to_obj(value: Any) -> Any:
            attr = getattr(value, "dict", None) or getattr(value, "dict", None)
            if attr and callable(attr):
                value = attr()
            return value

        return [to_obj(value) for value in self._value]

    def try_head(self) -> Option[_TSource]:
        """Returns the first element of the list, or None if the list is
        empty.
        """
        if self._value:
            value = cast("_TSource", self._value[0])
            return Some(value)

        return Nothing

    @staticmethod
    def unfold(
        generator: Callable[[_TState], Option[Tuple[_TSource, _TState]]], state: _TState
    ) -> Block[_TSource]:
        """Returns a list that contains the elements generated by the
        given computation. The given initial state argument is passed to
        the element generator.

        Args:
            generator: A function that takes in the current state and
                returns an option tuple of the next element of the list
                and the next state value.
            state: The initial state.

        Returns:
            The result list.
        """

        return pipe(state, unfold(generator))

    def zip(self, other: Block[_TResult]) -> Block[Tuple[_TSource, _TResult]]:
        """Combines the two lists into a list of pairs. The two lists
        must have equal lengths. .

        Args:
            other: The second input list.

        Returns:
            A single list containing pairs of matching elements from the
            input lists.
        """
        return of_seq(builtins.zip(self._value, other._value))

    def __add__(self, other: Block[_TSource]) -> Block[_TSource]:
        return Block(self._value + other._value)

    def __contains__(self, value: Any) -> bool:
        for v in self:
            if v is value or v == value:
                return True
        return False

    @overload
    def __getitem__(self, key: slice) -> Block[_TSource]:
        ...

    @overload
    def __getitem__(self, key: int) -> _TSource:
        ...

    def __getitem__(self, key: Any) -> Any:
        ret: Any = self._value[key]
        return ret

    def __iter__(self) -> Iterator[_TSource]:
        return iter(self._value)

    def __eq__(self, o: Any) -> bool:
        return self._value == o

    def __len__(self) -> int:
        return len(self._value)

    def __str__(self) -> str:
        return f"[{', '.join(self.map(str))}]"

    def __repr__(self) -> str:
        return str(self)

    @classmethod
    def __get_validators__(cls) -> Iterator[GenericValidator[Block[_TSource]]]:
        yield from cls.__validators__


@curry_flip(1)
def append(source: Block[_TSource], other: Block[_TSource]) -> Block[_TSource]:
    return source.append(other)


@curry_flip(1)
def choose(
    source: Block[_TSource], chooser: Callable[[_TSource], Option[_TResult]]
) -> Block[_TResult]:
    return source.choose(chooser)


@curry_flip(1)
def collect(
    source: Block[_TSource], mapping: Callable[[_TSource], Block[_TResult]]
) -> Block[_TResult]:
    """For each element of the list, applies the given function.
    Concatenates all the results and return the combined list.

    Args:
        source: The input list (curried flipped).
        mapping: The function to transform each input element into
        a sublist to be concatenated.

    Returns:
        A partially applied collect function that takes the source
        list and returns the concatenation of the transformed sublists.
    """

    return source.collect(mapping)


def concat(sources: Iterable[Block[_TSource]]) -> Block[_TSource]:
    """Concatenate sequence of Block's"""

    def reducer(t: Block[_TSource], s: Block[_TSource]) -> Block[_TSource]:
        return t.append(s)

    return pipe(sources, seq.fold(reducer, empty))


def cons(head: _TSource, tail: Block[_TSource]) -> Block[_TSource]:
    return Block((head, *tail))


nil: Block[Any] = Block()
empty: Block[Any] = nil
"""The empty list."""


@curry_flip(1)
def filter(
    source: Block[_TSource], predicate: Callable[[_TSource], bool]
) -> Block[_TSource]:
    """Returns a new collection containing only the elements of the
    collection for which the given predicate returns `True`

    Args:
        predicate: The function to test the input elements.

    Returns:
        Partially applied filter function.
    """

    return source.filter(predicate)


@curry_flip(1)
def fold(
    source: Block[_TSource],
    folder: Callable[[_TState, _TSource], _TState],
    state: _TState,
) -> _TState:
    """Applies a function to each element of the collection, threading
    an accumulator argument through the computation. Take the second
    argument, and apply the function to it and the first element of the
    list. Then feed this result into the function along with the second
    element and so on. Return the final result. If the input function is
    f and the elements are i0...iN then computes f (... (f s i0) i1 ...)
    iN.

    Args:
        folder: The function to update the state given the input
            elements.

        state: The initial state.

    Returns:
        Partially applied fold function that takes the source list
        and returns the final state value.
    """

    return source.fold(folder, state)


@curry_flip(1)
def forall(source: Block[_TSource], predicate: Callable[[_TSource], bool]) -> bool:
    """Tests if all elements of the collection satisfy the given
    predicate.

    Args:
        predicate: The function to test the input elements.

    Returns:
        True if all of the elements satisfy the predicate.
    """

    return source.forall(predicate)


def head(source: Block[_TSource]) -> _TSource:
    """Returns the first element of the list.

    Args:
        source: The input list.

    Returns:
        The first element of the list.

    Raises:
         ValueError: Thrown when the list is empty.
    """
    return source.head()


def indexed(source: Block[_TSource]) -> Block[Tuple[int, _TSource]]:
    """Returns a new list whose elements are the corresponding
    elements of the input list paired with the index (from 0)
    of each element.

    Returns:
        The list of indexed elements.
    """
    return source.indexed()


@curry_flip(1)
def item(source: Block[_TSource], index: int) -> _TSource:
    """Indexes into the list. The first element has index 0.

    Args:
        index: The index to retrieve.

    Returns:
        The value at the given index.
    """

    return source.item(index)


def is_empty(source: Block[Any]) -> bool:
    """Returns `True` if the list is empty, `False` otherwise."""
    return source.is_empty()


@curry_flip(1)
def map(
    source: Block[_TSource], mapper: Callable[[_TSource], _TResult]
) -> Block[_TResult]:
    """Map list.

    Builds a new collection whose elements are the results of applying
    the given function to each of the elements of the collection.

    Args:
        mapper: The function to transform elements from the input list.

    Returns:
        The list of transformed elements.
    """

    return source.map(mapper)


@curry_flip(1)
def reduce(
    source: Block[_TSource],
    reduction: Callable[[_TSource, _TSource], _TSource],
) -> _TSource:
    """Apply a function to each element of the collection, threading an
    accumulator argument through the computation. Apply the function to
    the first two elements of the list. Then feed this result into the
    function along with the third element and so on. Return the final
    result. If the input function is f and the elements are i0...iN then
    computes f (... (f i0 i1) i2 ...) iN.

    Args:
        source: The input block (curried flipped)
        reduction: The function to reduce two list elements to a single
            element.

    Returns:
        Partially applied reduce function that takes the source block
        and returns the final state value.
    """

    if source.is_empty():
        raise ValueError("Collection was empty")

    return source.tail().fold(reduction, source.head())


@overload
def starmap(
    mapper: Callable[[_T1, _T2], _TResult]
) -> Callable[[Block[Tuple[_T1, _T2]]], Block[_TResult]]:
    ...


@overload
def starmap(
    mapper: Callable[[_T1, _T2, _T3], _TResult]
) -> Callable[[Block[Tuple[_T1, _T2, _T3]]], Block[_TResult]]:
    ...


@overload
def starmap(
    mapper: Callable[[_T1, _T2, _T3, _T4], _TResult]
) -> Callable[[Block[Tuple[_T1, _T2, _T3, _T4]]], Block[_TResult]]:
    ...


def starmap(
    mapper: Callable[..., Any]
) -> Callable[[Block[Tuple[Any, ...]]], Block[Any]]:
    """Starmap source sequence.

    Unpack arguments grouped as tuple elements. Builds a new collection
    whose elements are the results of applying the given function to the
    unpacked arguments to each of the elements of the collection.

    Args:
        mapping: A function to transform items from the input sequence.

    Returns:
        Partially applied map function.
    """

    def mapper_(args: Tuple[Any, ...]) -> Any:
        return mapper(*args)

    return map(mapper_)


def map2(
    mapper: Callable[[_T1, _T2], _TResult]
) -> Callable[[Block[Tuple[_T1, _T2]]], Block[_TResult]]:
    return starmap(mapper)


def map3(
    mapper: Callable[[_T1, _T2, _T3], _TResult]
) -> Callable[[Block[Tuple[_T1, _T2, _T3]]], Block[_TResult]]:
    return starmap(mapper)


@curry_flip(1)
def mapi(
    source: Block[_TSource], mapper: Callable[[int, _TSource], _TResult]
) -> Block[_TResult]:
    """Map list with index.

    Builds a new collection whose elements are the results of
    applying the given function to each of the elements of the
    collection. The integer index passed to the function indicates
    the index (from 0) of element being transformed.

    Args:
        mapping: The function to transform elements and their
            indices.

    Returns:
        The list of transformed elements.
    """

    return source.mapi(mapper)


def of(*args: _TSource) -> Block[_TSource]:
    """Create list from a number of arguments."""
    return Block((*args,))


def of_seq(xs: Iterable[_TSource]) -> Block[_TSource]:
    """Create list from iterable sequence."""
    return Block((*xs,))


def of_option(option: Option[_TSource]) -> Block[_TSource]:
    if isinstance(option, Some):
        return singleton(option.value)
    return empty


@curry_flip(1)
def partition(
    source: Block[_TSource], predicate: Callable[[_TSource], bool]
) -> Tuple[Block[_TSource], Block[_TSource]]:
    """Partition block.

    Splits the collection into two collections, containing the
    elements for which the given predicate returns True and False
    respectively. Element order is preserved in both of the created
    lists.

    Args:
        source: The source block to partition (curried flipped)
        predicate: The function to test the input elements.

    Returns:
        A list containing the elements for which the predicate
        evaluated to true and a list containing the elements for
        which the predicate evaluated to false.
    """
    return source.partition(predicate)


@overload
def range(stop: int) -> Block[int]:
    ...


@overload
def range(start: int, stop: int) -> Block[int]:
    ...


@overload
def range(start: int, stop: int, step: int) -> Block[int]:
    ...


def range(*args: int, **kw: int) -> Block[int]:
    return Block((*builtins.range(*args, **kw),))


def singleton(value: _TSource) -> Block[_TSource]:
    return Block((value,))


@curry_flip(1)
def skip(source: Block[_TSource], count: int) -> Block[_TSource]:
    """Returns the list after removing the first N elements.

    Args:
        count: The number of elements to skip.

    Returns:
        The list after removing the first N elements.
    """

    return source.skip(count)


@curry_flip(1)
def skip_last(source: Block[_TSource], count: int) -> Block[_TSource]:
    """Returns the list after removing the last N elements.

    Args:
        count: The number of elements to skip.

    Returns:
        The list after removing the last N elements.
    """

    return source.skip_last(count)


@curry_flip(1)
def sort(
    source: Block[_TSourceSortable],
    reverse: bool = False,
) -> Block[_TSourceSortable]:
    """Returns a new sorted collection

    Args:
        reverse: Sort in reversed order.

    Returns:
        Partially applied sort function.
    """
    return source.sort(reverse)


@curry_flip(1)
def sort_with(
    source: Block[_TSource], func: Callable[[_TSource], Any], reverse: bool = False
) -> Block[_TSource]:
    """Returns a new collection sorted using "func" key function.

    Args:
        func: The function to extract a comparison key from each element in list.
        reverse: Sort in reversed order.

    Returns:
        Partially applied sort function.
    """

    return source.sort_with(func, reverse)


def sum(source: Block[_TSourceSum | Literal[0]]) -> _TSourceSum | Literal[0]:
    return builtins.sum(source)


@curry_flip(1)
def sum_by(
    source: Block[_TSourceSum], projection: Callable[[_TSourceSum], _TResult]
) -> _TResult:
    xs = source.map(projection)
    return builtins.sum(xs)  # type: ignore


def tail(source: Block[_TSource]) -> Block[_TSource]:
    return source.tail()


@curry_flip(1)
def take(source: Block[_TSource], count: int) -> Block[_TSource]:
    """Returns the first N elements of the list.

    Args:
        count: The number of items to take.

    Returns:
        The result list.
    """
    return source.take(count)


@curry_flip(1)
def take_last(source: Block[_TSource], count: int) -> Block[_TSource]:
    """Returns a specified number of contiguous elements from the end of
    the list.

    Args:
        count: The number of items to take.

    Returns:
        The result list.
    """

    return source.take_last(count)


def dict(source: Block[_TSource]) -> List[_TSource]:
    return source.dict()


def try_head(source: Block[_TSource]) -> Option[_TSource]:
    """Try to get the first element from the list.

    Returns the first element of the list, or None if the list is empty.

    Args:
        source: The input list.

    Returns:
        The first element of the list or `Nothing`.
    """
    return source.try_head()


@curry_flip(1)
def unfold(
    state: _TState, generator: Callable[[_TState], Option[Tuple[_TSource, _TState]]]
) -> Block[_TSource]:
    """Returns a list that contains the elements generated by the
    given computation. The given initial state argument is passed to
    the element generator.

    Args:
        generator: A function that takes in the current state and
            returns an option tuple of the next element of the list
            and the next state value.
        state: The initial state.

    Returns:
        The result list.
    """

    xs = pipe(state, seq.unfold(generator))
    return Block(xs)


@curry_flip(1)
def zip(
    source: Block[_TSource],
    other: Block[_TResult],
) -> Block[Tuple[_TSource, _TResult]]:
    """Combines the two lists into a list of pairs. The two lists
    must have equal lengths.

    Args:
        other: The second input list.

    Returns:
        Paritally applied zip function that takes the source list and
        returns s single list containing pairs of matching elements from
        the input lists.
    """

    return source.zip(other)


__all__ = [
    "Block",
    "append",
    "choose",
    "collect",
    "concat",
    "dict",
    "empty",
    "filter",
    "fold",
    "head",
    "indexed",
    "item",
    "is_empty",
    "map",
    "mapi",
    "of_seq",
    "of_option",
    "partition",
    "singleton",
    "skip",
    "skip_last",
    "sort",
    "sort_with",
    "tail",
    "take",
    "take_last",
    "try_head",
    "unfold",
    "zip",
]



"""The `Try` type  is a simpler `Result` type that pins the error type
to Exception.

Everything else is the same as `Result`, just simpler to use.
"""

from typing import TypeVar

from .result import Error, Ok, Result

_TSource = TypeVar("_TSource")

Try = Result[_TSource, Exception]


class Success(Ok[_TSource, Exception]):
    """The successful Try case.

    Same as result `Ok` but with error type pinned to an exception, i.e:
    `Ok[TSource, Exception]`
    """

    ...


class Failure(Error[_TSource, Exception]):
    """The failure Try case.

    Same as result `Error` but with error type pinned to an exception,
    i.e: `Error[TSource, Exception]`
    """

    ...



"""Result module.

The Result[TSource,TError] type lets you write error-tolerant code that
can be composed. The Result type is typically used in monadic
error-handling, which is often referred to as Railway-oriented
Programming.

There is also a simplifyed alias of this type called `Maybe` that pins
the Result type to Exception.
"""
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    Generic,
    Iterable,
    Iterator,
    List,
    TypeVar,
    Union,
    cast,
    get_origin,
)

from typing_extensions import TypeGuard

from .curry import curry_flip
from .error import EffectError
from .pipe import PipeMixin
from .typing import GenericValidator, ModelField, SupportsValidation

_TSource = TypeVar("_TSource")
_TOther = TypeVar("_TOther")
_TResult = TypeVar("_TResult")
_TError = TypeVar("_TError")


def _validate(result: Any, field: ModelField) -> Result[Any, Any]:
    if isinstance(result, Result):
        return cast(Result[Any, Any], result)

    if not isinstance(result, Dict):
        raise ValueError("not result type")

    try:
        value: Any = result["ok"]
        if field.sub_fields:
            sub_field = field.sub_fields[0]
            value, err = sub_field.validate(value, {}, loc="Result")
            if err:
                raise ValueError(str(err))
        return Ok(value)
    except KeyError:
        try:
            error: Any = result["error"]
            sub_field = field.sub_fields[1]
            error, err = sub_field.validate(error, {}, loc="Result")
            if err:
                raise ValueError(str(err))
            return Error(error)
        except KeyError:
            raise ValueError("not a result")


class Result(
    Iterable[_TSource],
    PipeMixin,
    SupportsValidation["Result[_TSource, _TError]"],
    Generic[_TSource, _TError],
    ABC,
):
    """The result abstract base class."""

    __validators__: List[GenericValidator[Result[_TSource, _TError]]] = [_validate]

    @abstractmethod
    def default_value(self, value: _TSource) -> _TSource:
        """Get with default value.

        Gets the value of the option if the option is Some, otherwise
        returns the specified default value.
        """
        raise NotImplementedError

    @abstractmethod
    def default_with(self, getter: Callable[[_TError], _TSource]) -> _TSource:
        """Get with default value lazily.

        Gets the value of the option if the option is Some, otherwise
        returns the value produced by the getter
        """
        raise NotImplementedError

    @abstractmethod
    def map(self, mapper: Callable[[_TSource], _TResult]) -> Result[_TResult, _TError]:
        raise NotImplementedError

    @abstractmethod
    def map2(
        self,
        other: Ok[_TOther, _TError] | Error[_TOther, _TError],
        mapper: Callable[[_TSource, _TOther], _TResult],
    ) -> Result[_TResult, _TError]:
        raise NotImplementedError

    @abstractmethod
    def map_error(
        self, mapper: Callable[[_TError], _TResult]
    ) -> Result[_TSource, _TResult]:
        """Return a result of the error value after applying the mapping
        function, or Ok if the input is Ok."""
        raise NotImplementedError

    @abstractmethod
    def bind(
        self, mapper: Callable[[_TSource], Result[_TResult, _TError]]
    ) -> Result[_TResult, _TError]:
        raise NotImplementedError

    @abstractmethod
    def is_error(self) -> bool:
        """Returns `True` if the result is an `Error` value."""

        raise NotImplementedError

    @abstractmethod
    def is_ok(self) -> bool:
        """Returns `True` if the result is an `Ok` value."""

        raise NotImplementedError

    @abstractmethod
    def dict(self) -> Dict[str, Union[_TSource, _TError]]:
        """Returns a json serializable representation of the result."""
        raise NotImplementedError

    def __eq__(self, o: Any) -> bool:
        raise NotImplementedError

    @abstractmethod
    def __iter__(self) -> Generator[_TSource, _TSource, _TSource]:
        raise NotImplementedError

    def __repr__(self) -> str:
        return str(self)

    @classmethod
    def __get_validators__(
        cls,
    ) -> Iterator[GenericValidator[Result[_TSource, _TError]]]:
        yield from cls.__validators__

    @abstractmethod
    def __hash__(self) -> int:
        raise NotImplementedError


class Ok(Result[_TSource, _TError]):
    """The Ok result case class."""

    __match_args__ = ("value",)

    def default_value(self, value: _TSource) -> _TSource:
        """Get with default value.

        Gets the value of the option if the option is Some, otherwise
        returns the specified default value.
        """
        return self._value

    def default_with(self, getter: Callable[[_TError], _TSource]) -> _TSource:
        """Get with default value lazily.

        Gets the value of the option if the option is Some, otherwise
        returns the value produced by the getter
        """
        return self._value

    def __init__(self, value: _TSource) -> None:
        self._value = value

    @property
    def value(self) -> _TSource:
        return self._value

    def map(self, mapper: Callable[[_TSource], _TResult]) -> Result[_TResult, _TError]:
        return Ok(mapper(self._value))

    def map2(
        self,
        other: Ok[_TOther, _TError] | Error[_TOther, _TError],
        mapper: Callable[[_TSource, _TOther], _TResult],
    ) -> Result[_TResult, _TError]:
        return other.map(lambda value: mapper(self._value, value))

    def bind(
        self, mapper: Callable[[_TSource], Result[_TResult, _TError]]
    ) -> Result[_TResult, _TError]:
        return mapper(self._value)

    def map_error(
        self, mapper: Callable[[_TError], _TResult]
    ) -> Result[_TSource, _TResult]:
        """Return a result of the error value after applying the mapping
        function, or Ok if the input is Ok."""
        return Ok(self._value)

    def is_error(self) -> bool:
        """Returns `True` if the result is an `Ok` value."""

        return False

    def is_ok(self) -> bool:
        """Returns `True` if the result is an `Ok` value."""

        return True

    def dict(self) -> Dict[str, _TSource | _TError]:
        """Returns a json string representation of the ok value."""
        attr = getattr(self._value, "dict", None) or getattr(self._value, "dict", None)
        if attr and callable(attr):
            value = attr()
        else:
            value = self._value

        return {"ok": value}

    def __match__(self, pattern: Any) -> Iterable[_TSource]:
        if self is pattern or self == pattern:
            return [self.value]

        try:
            origin: Any = get_origin(pattern)
            if isinstance(self, origin or pattern):
                return [self.value]
        except TypeError:
            pass

        return []

    def __eq__(self, o: Any) -> bool:
        if isinstance(o, Ok):
            return self.value == o.value  # type: ignore
        return False

    def __iter__(self) -> Generator[_TSource, _TSource, _TSource]:
        """Return iterator for Ok case."""
        return (yield self._value)

    def __str__(self):
        return f"Ok {self._value}"

    def __hash__(self) -> int:
        return hash(self._value)


class ResultException(EffectError):
    """Makes the Error case a valid exception for effect handling. Do
    not use directly."""

    def __init__(self, message: str):
        self.message = message


class Error(
    ResultException,
    Result[_TSource, _TError],
):
    """The Error result case class."""

    __match_args__ = ("error",)

    def __init__(self, error: _TError) -> None:
        super().__init__(str(error))
        self._error = error

    def default_value(self, value: _TSource) -> _TSource:
        """Get with default value.

        Gets the value of the option if the option is Some, otherwise
        returns the specified default value.
        """
        return value

    def default_with(self, getter: Callable[[_TError], _TSource]) -> _TSource:
        """Get with default value lazily.

        Gets the value of the option if the option is Some, otherwise
        returns the value produced by the getter
        """
        return getter(self._error)

    @property
    def error(self) -> _TError:
        return self._error

    def map(self, mapper: Callable[[_TSource], _TResult]) -> Result[_TResult, _TError]:
        return Error(self._error)

    def map2(
        self,
        other: Result[_TOther, _TError],
        mapper: Callable[[_TSource, _TOther], _TResult],
    ) -> Result[_TResult, _TError]:
        return Error(self._error)

    def bind(
        self, mapper: Callable[[_TSource], Result[_TResult, _TError]]
    ) -> Result[_TResult, _TError]:
        return Error(self._error)

    def map_error(
        self, mapper: Callable[[_TError], _TResult]
    ) -> Result[_TSource, _TResult]:
        """Return a result of the error value after applying the mapping
        function, or Ok if the input is Ok."""
        return Error(mapper(self._error))

    def is_error(self) -> bool:
        """Returns `True` if the result is an `Ok` value."""
        return True

    def is_ok(self) -> bool:
        """Returns `True` if the result is an `Ok` value."""
        return False

    def dict(self) -> Dict[str, Any]:
        """Returns a json serializable representation of the error value."""
        attr = getattr(self._error, "dict") or getattr(self._error, "dict")
        if callable(attr):
            error = attr()
        else:
            error = self._error

        return {"error": error}

    def __eq__(self, o: Any) -> bool:
        if isinstance(o, Error):
            return self.error == o.error  # type: ignore
        return False

    def __iter__(self) -> Generator[_TSource, _TSource, _TSource]:
        """Return iterator for Error case."""

        # Raise class here so sub-classes like Failure works as well.
        raise self.__class__(self._error)

        # We're a generator
        while False:
            yield

    def __str__(self):
        return f"Error {self._error}"

    def __hash__(self) -> int:
        return hash(self._error)


def default_value(value: _TSource) -> Callable[[Result[_TSource, Any]], _TSource]:
    """Gets the value of the option if the option is Some, otherwise
    returns the specified default value.
    """

    def _default_value(result: Result[_TSource, Any]) -> _TSource:
        return result.default_value(value)

    return _default_value


def default_with(
    getter: Callable[[_TError], _TSource]
) -> Callable[[Result[_TSource, _TError]], _TSource]:
    """Get with default value lazily.

    Gets the value of the option if the option is Some, otherwise
    returns the value produced by the getter
    """

    def _default_with(result: Result[_TSource, _TError]) -> _TSource:
        return result.default_with(getter)

    return _default_with


@curry_flip(1)
def map(
    result: Result[_TSource, _TError], mapper: Callable[[_TSource], _TResult]
) -> Result[_TResult, _TError]:
    return result.map(mapper)


@curry_flip(2)
def map2(
    x: Ok[_TSource, _TError] | Error[_TSource, _TError],
    y: Ok[_TOther, _TError] | Error[_TOther, _TError],
    mapper: Callable[[_TSource, _TOther], _TResult],
) -> Result[_TResult, _TError]:
    return x.map2(y, mapper)


@curry_flip(1)
def bind(
    result: Result[_TSource, _TError],
    mapper: Callable[[_TSource], Result[_TResult, Any]],
) -> Result[_TResult, _TError]:
    return result.bind(mapper)


def dict(source: Result[_TSource, _TError]) -> Dict[str, Union[_TSource, _TError]]:
    return source.dict()


def is_ok(result: Result[_TSource, _TError]) -> TypeGuard[Ok[_TSource, _TError]]:
    """Returns `True` if the result is an `Ok` value."""

    return result.is_ok()


def is_error(result: Result[_TSource, _TError]) -> TypeGuard[Error[_TSource, _TError]]:
    """Returns `True` if the result is an `Error` value."""

    return result.is_error()


__all__ = [
    "Result",
    "Ok",
    "Error",
    "default_value",
    "default_with",
    "map",
    "bind",
    "dict",
    "is_ok",
    "is_error",
]



"""Choice type.

A union type similar to the `Result` type. But also allows for higher
number of choices. Usually you would most likely want to use the
`Result` type instead, but choice can be preferred in non-error cases.
"""
from abc import ABC
from typing import Any, Generic, TypeVar

_TSource = TypeVar("_TSource")
_A = TypeVar("_A")
_B = TypeVar("_B")
_C = TypeVar("_C")


class Choice(ABC, Generic[_TSource]):
    def __init__(self, value: Any) -> None:
        self.value = value

    def __repr__(self) -> str:
        return str(self)

    def __hash__(self) -> int:
        return hash(self.value)


class Choice2(Generic[_A, _B]):
    def __repr__(self) -> str:
        return str(self)


class Choice1of2(Choice2[_A, _B], Choice[_A]):
    __match_args__ = ("value",)

    def __init__(self, value: _A) -> None:
        super().__init__(value)

    def __eq__(self, o: Any) -> bool:
        if isinstance(o, Choice1of2):
            return self.value == o.value
        return False

    def __str__(self) -> str:
        return f"Choice1of2 {self.value}"


class Choice2of2(Choice2[_A, _B], Choice[_B]):
    __match_args__ = ("value",)

    def __init__(self, value: _B) -> None:
        super().__init__(value)

    def __eq__(self, o: Any) -> bool:
        if isinstance(o, Choice2of2):
            return self.value == o.value
        return False

    def __str__(self) -> str:
        return f"Choice2of2 {self.value}"


class Choice3(Generic[_A, _B, _C]):
    ...


class Choice1of3(Choice3[_A, _B, _C], Choice[_A]):
    __match_args__ = ("value",)

    def __init__(self, value: _A) -> None:
        super().__init__(value)


class Choice2of3(Choice3[_A, _B, _C], Choice[_B]):
    __match_args__ = ("value",)

    def __init__(self, value: _B) -> None:
        super().__init__(value)


class Choice3of3(Choice3[_A, _B, _C], Choice[_C]):
    __match_args__ = ("value",)

    def __init__(self, value: _C) -> None:
        super().__init__(value)


__all__ = [
    "Choice",
    "Choice2",
    "Choice3",
    "Choice1of2",
    "Choice2of2",
    "Choice1of3",
    "Choice2of3",
    "Choice3of3",
]



# Attribution to original authors of this code
# --------------------------------------------
# This code has been originally been ported from the Fable project which
# was originally ported from the FSharp project.
#
# Fable:
#   https://fable.io
# - Copyright (c) Alfonso Garcia-Caro and contributors.
# - MIT License
# - https://github.com/fable-compiler/Fable/blob/nagareyama/src/fable-library/Map.fs
#
# F#
# - https://github.com/dotnet/fsharp
# - Copyright (c) Microsoft Corporation. All Rights Reserved.
# - MIT License
# - https://github.com/fsharp/fsharp/blob/master/src/fsharp/FSharp.Core/map.fs

"""
The maptree module.

Contains the internal tree implementation of the `map`.

Do not use directly. Use the `map` module instead.
"""
import builtins
from dataclasses import dataclass
from typing import Any, Callable, Generic, Iterable, Iterator, Tuple, TypeVar, cast

from expression.core import Nothing, Option, Some, SupportsLessThan, failwith, pipe

from . import block, seq
from .block import Block

Key = TypeVar("Key", bound=SupportsLessThan)
Value = TypeVar("Value")
Result = TypeVar("Result")


@dataclass
class MapTreeLeaf(Generic[Key, Value]):
    key: Key
    value: Value


MapTree = Option[MapTreeLeaf[Key, Value]]


@dataclass
class MapTreeNode(MapTreeLeaf[Key, Value]):
    left: MapTree[Key, Value]
    right: MapTree[Key, Value]

    height: int


empty: MapTree[Any, Any] = Nothing


def is_empty(m: MapTree[Any, Any]):
    return m.is_none()


def size_aux(acc: int, m: MapTree[Any, Any]) -> int:
    for m2 in m.to_list():
        if isinstance(m2, MapTreeNode):
            mn = cast(MapTreeNode[Key, Value], m2)
            return size_aux(size_aux(acc + 1, mn.left), mn.right)
        else:
            return acc + 1
    else:
        return acc


def size(x: MapTree[Any, Any]):
    return size_aux(0, x)


def height(m: MapTree[Any, Any]) -> int:
    for m2 in m.to_list():
        if isinstance(m2, MapTreeNode):
            mn = cast(MapTreeNode[Key, Value], m2)
            return mn.height
        else:
            return 1
    else:
        return 0


TOLERANCE = 2


def mk(
    left: MapTree[Key, Value], key: Key, value: Value, right: MapTree[Key, Value]
) -> MapTree[Key, Value]:
    hl = height(left)
    hr = height(right)
    m = hr if hl < hr else hl
    if m == 0:  # m=0 ~ is_empty(l) and is_empty(r)
        return Some(MapTreeLeaf(key, value))
    else:
        return Some(
            MapTreeNode(key, value, left, right, m + 1)
        )  # new map is higher by 1 than the highest


def rebalance(
    t1: MapTree[Key, Value], k: Key, v: Value, t2: MapTree[Key, Value]
) -> MapTree[Key, Value]:
    t1h = height(t1)
    t2h = height(t2)
    if t2h > t1h + TOLERANCE:  # right is heavier than left
        if isinstance(t2.value, MapTreeNode):
            t2_ = t2.value
            # One of the nodes must have height > height t1 + 1
            if height(t2_.left) > t1h + 1:  # balance left: combination
                if isinstance(t2_.left.value, MapTreeNode):
                    t2l = t2_.left.value
                    return mk(
                        mk(t1, k, v, t2l.left),
                        t2l.key,
                        t2l.value,
                        mk(t2l.right, t2_.key, t2_.value, t2_.right),
                    )
                else:
                    failwith("internal error: Map.rebalance")
            else:  # Rotate left
                return mk(mk(t1, k, v, t2_.left), t2_.key, t2_.value, t2_.right)
        else:
            failwith("internal error: Map.rebalance")
    else:
        if t1h > t2h + TOLERANCE:  # left is heavier than right
            if isinstance(t1.value, MapTreeNode):
                t1_ = t1.value
                # One of the nodes must have height > height t2 + 1
                if height(t1_.right) > t2h + 1:  # balance right: combination
                    if isinstance(t1_.right.value, MapTreeNode):
                        t1r = t1_.right.value
                        return mk(
                            mk(t1_.left, t1_.key, t1_.value, t1r.left),
                            t1r.key,
                            t1r.value,
                            mk(t1r.right, k, v, t2),
                        )
                    else:
                        failwith("internal error: Map.rebalance")
                else:
                    return mk(t1_.left, t1_.key, t1_.value, mk(t1_.right, k, v, t2))
            else:
                failwith("internal error: Map.rebalance")
        else:
            return mk(t1, k, v, t2)


def add(k: Key, v: Value, m: MapTree[Key, Value]) -> MapTree[Key, Value]:
    for m2 in m.to_list():
        if isinstance(m2, MapTreeNode):
            mn = m2
            if k < mn.key:
                return rebalance(add(k, v, mn.left), mn.key, mn.value, mn.right)
            elif k == mn.key:
                return Some(MapTreeNode(k, v, mn.left, mn.right, mn.height))
            else:
                return rebalance(mn.left, mn.key, mn.value, add(k, v, mn.right))
        else:
            if k < m2.key:
                node = MapTreeNode(k, v, empty, m, 2)
                return Some(node)
            elif k == m2.key:
                return Some(MapTreeLeaf(k, v))
            else:
                return Some(MapTreeNode(k, v, m, empty, 2))
    else:
        return Some(MapTreeLeaf(k, v))


def try_find(k: Key, m: MapTree[Key, Value]) -> Option[Value]:
    for m2 in m.to_list():
        if k == m2.key:
            return Some(m2.value)
        else:
            if isinstance(m2, MapTreeNode):
                mn = m2
                return try_find(k, mn.left if k < mn.key else mn.right)
            else:
                return Nothing
    else:  # Nothing
        return Nothing


def find(k: Key, m: MapTree[Key, Value]) -> Value:
    for v in try_find(k, m).to_list():
        return v
    else:
        raise KeyError("Key not found")


def partition1(
    predicate: Callable[[Key, Value], bool],
    k: Key,
    v: Value,
    acc: Tuple[MapTree[Key, Value], MapTree[Key, Value]],
) -> Tuple[MapTree[Key, Value], MapTree[Key, Value]]:
    acc1, acc2 = acc
    if predicate(k, v):
        a: MapTree[Key, Value] = add(k, v, acc1)
        return a, acc2
    else:
        return acc1, add(k, v, acc2)


def partition_aux(
    predicate: Callable[[Key, Value], bool],
    m: MapTree[Key, Value],
    acc: Tuple[MapTree[Key, Value], MapTree[Key, Value]],
) -> Tuple[MapTree[Key, Value], MapTree[Key, Value]]:
    for m2 in m:
        if isinstance(m2, MapTreeNode):
            mn = m2
            acc = partition_aux(predicate, mn.right, acc)
            acc = partition1(predicate, mn.key, mn.value, acc)
            return partition_aux(predicate, mn.left, acc)
        else:
            return partition1(predicate, m2.key, m2.value, acc)
    else:  # Nothing
        return acc


def partition(
    predicate: Callable[[Key, Value], bool], m: MapTree[Key, Value]
) -> Tuple[MapTree[Key, Value], MapTree[Key, Value]]:
    return partition_aux(predicate, m, (empty, empty))


def filter1(
    predicate: Callable[[Key, Value], bool], k: Key, v: Value, acc: MapTree[Key, Value]
) -> MapTree[Key, Value]:
    if predicate(k, v):
        return add(k, v, acc)
    else:
        return acc


def filter_aux(
    predicate: Callable[[Key, Value], bool],
    m: MapTree[Key, Value],
    acc: MapTree[Key, Value],
) -> MapTree[Key, Value]:
    for m2 in m.to_list():
        if isinstance(m2, MapTreeNode):
            mn = m2
            acc = filter_aux(predicate, mn.left, acc)
            acc = filter1(predicate, mn.key, mn.value, acc)
            return filter_aux(predicate, mn.right, acc)
        else:
            return filter1(predicate, m2.key, m2.value, acc)
    else:  # Nothing
        return acc


def filter(
    f: Callable[[Key, Value], bool], m: MapTree[Key, Value]
) -> MapTree[Key, Value]:
    return filter_aux(f, m, empty)


def splice_out_successor(
    m: MapTree[Key, Value]
) -> Tuple[Key, Value, Option[MapTreeLeaf[Key, Value]]]:
    for m2 in m.to_list():
        if isinstance(m2, MapTreeNode):
            mn = m2
            if is_empty(mn.left):
                return mn.key, mn.value, mn.right
            else:
                k3, v3, l_ = splice_out_successor(mn.left)
                return k3, v3, mk(l_, mn.key, mn.value, mn.right)
        else:
            return m2.key, m2.value, empty
    else:  # Nothing
        failwith("internal error: Map.splice_out_successor")


def remove(k: Key, m: MapTree[Key, Value]) -> Option[MapTreeLeaf[Key, Value]]:
    for m2 in m.to_list():
        if isinstance(m2, MapTreeNode):
            mn = m2
            if k < mn.key:
                return rebalance(remove(k, mn.left), mn.key, mn.value, mn.right)
            elif k == mn.key:
                if is_empty(mn.left):
                    return mn.right
                elif is_empty(mn.right):
                    return mn.left
                else:
                    sk, sv, r_ = splice_out_successor(mn.right)
                    return mk(mn.left, sk, sv, r_)
            else:
                return rebalance(mn.left, mn.key, mn.value, remove(k, mn.right))
        else:
            if k == m2.key:
                return empty
            else:
                return m
    else:  # Nothing
        return empty


def change(
    k: Key, u: Callable[[Option[Value]], Option[Value]], m: MapTree[Key, Value]
) -> MapTree[Key, Value]:
    for m2 in m.to_list():
        if isinstance(m2, MapTreeNode):
            mn = m2
            if k < mn.key:
                return rebalance(change(k, u, mn.left), mn.key, mn.value, mn.right)
            elif k == mn.key:
                for v in u(Some(mn.value)).to_list():
                    return Some(MapTreeNode(k, v, mn.left, mn.right, mn.height))
                else:
                    if is_empty(mn.left):
                        return mn.right
                    elif is_empty(mn.right):
                        return mn.left
                    else:
                        sk, sv, r_ = splice_out_successor(mn.right)
                        return mk(mn.left, sk, sv, r_)
            else:
                return rebalance(mn.left, mn.key, mn.value, change(k, u, mn.right))
        else:
            if k < m2.key:
                for v in u(Nothing).to_list():
                    return Some(MapTreeNode(k, v, empty, m, 2))
                else:
                    return m
            elif k == m2.key:
                for v in u(Some(m2.value)).to_list():
                    return Some(MapTreeLeaf(k, v))
                else:
                    return empty
            else:
                for v in u(Nothing).to_list():
                    return Some(MapTreeNode(k, v, m, empty, 2))
                else:
                    return m

    else:
        for v in u(Nothing):
            return Some(MapTreeLeaf(k, v))
        else:
            return m


def mem(k: Key, m: MapTree[Key, Any]) -> bool:
    for m2 in m.to_list():
        if isinstance(m2, MapTreeNode):
            mn = cast(MapTreeNode[Key, Value], m2)
            if k < mn.key:
                return mem(k, mn.left)
            else:
                return k == mn.key or mem(k, mn.right)
        else:
            return k == m2.key
    else:
        return False


def iter(fn: Callable[[Key, Value], None], m: MapTree[Key, Value]) -> None:
    """Iterate maptree."""
    for m2 in m.to_list():
        if isinstance(m2, MapTreeNode):
            mn = m2
            iter(fn, mn.left)
            fn(mn.key, mn.value)
            iter(fn, mn.right)
        else:
            fn(m2.key, m2.value)


def try_pick(
    f: Callable[[Key, Value], Option[Result]], m: MapTree[Key, Value]
) -> Option[Result]:
    for m2 in m.to_list():
        if isinstance(m2, MapTreeNode):
            mn = m2
            res = try_pick(f, mn.left)
            if res.is_some():
                return res
            else:
                res = f(mn.key, mn.value)
                if res.is_some():
                    return res
                else:
                    return try_pick(f, mn.right)
        else:
            return f(m2.key, m2.value)
    else:
        return Nothing


def exists(f: Callable[[Key, Value], bool], m: MapTree[Key, Value]) -> bool:
    for m2 in m.to_list():
        if isinstance(m2, MapTreeNode):
            mn = m2
            return exists(f, mn.left) or f(mn.key, mn.value) or exists(f, mn.right)
        else:
            return f(m2.key, m2.value)
    else:
        return False


def forall(f: Callable[[Key, Value], bool], m: MapTree[Key, Value]) -> bool:
    for m2 in m.to_list():
        if isinstance(m2, MapTreeNode):
            mn = m2
            return forall(f, mn.left) and f(mn.key, mn.value) and forall(f, mn.right)
        else:
            return f(m2.key, m2.value)
    else:
        return True


def map(
    f: Callable[[Key, Value], Result], m: MapTree[Key, Value]
) -> MapTree[Key, Result]:
    for m2 in m.to_list():
        if isinstance(m2, MapTreeNode):
            mn = m2
            l2 = map(f, mn.left)
            v2 = f(mn.key, mn.value)
            r2 = map(f, mn.right)
            return Some(MapTreeNode(mn.key, v2, l2, r2, mn.height))
        else:
            return Some(MapTreeLeaf(m2.key, f(m2.key, m2.value)))
    else:
        return empty


def fold_back(
    f: Callable[[Tuple[Key, Value], Result], Result], m: MapTree[Key, Value], x: Result
) -> Result:
    for m2 in m.to_list():
        if isinstance(m2, MapTreeNode):
            mn = m2
            x = fold_back(f, mn.right, x)
            x = f((mn.key, mn.value), x)
            return fold_back(f, mn.left, x)
        else:
            return f((m2.key, m2.value), x)
    else:
        return x


def fold(
    f: Callable[[Result, Tuple[Key, Value]], Result], x: Result, m: MapTree[Key, Value]
) -> Result:
    for m2 in m.to_list():
        if isinstance(m2, MapTreeNode):
            mn = m2
            x = fold(f, x, mn.left)
            x = f(x, (mn.key, mn.value))
            return fold(f, x, mn.right)
        else:
            return f(x, (m2.key, m2.value))
    else:
        return x


def to_list(m: MapTree[Key, Value]) -> Block[Tuple[Key, Value]]:
    def loop(
        m: MapTree[Key, Value], acc: Block[Tuple[Key, Value]]
    ) -> Block[Tuple[Key, Value]]:
        for m2 in m.to_list():
            if isinstance(m2, MapTreeNode):
                mn = m2
                return loop(mn.left, loop(mn.right, acc).cons((mn.key, mn.value)))
            else:
                return acc.cons((m2.key, m2.value))
        else:
            return acc

    return loop(m, block.empty)


def of_list(xs: Block[Tuple[Key, Value]]) -> MapTree[Key, Value]:
    def folder(acc: MapTree[Key, Value], kv: Tuple[Key, Value]):
        k, v = kv
        return add(k, v, acc)

    return xs.fold(folder, empty)


def mk_from_iterator(
    acc: MapTree[Key, Value], e: Iterator[Tuple[Key, Value]]
) -> MapTree[Key, Value]:
    try:
        (x, y) = next(e)
    except StopIteration:
        return acc
    else:
        return mk_from_iterator(add(x, y, acc), e)


def of_seq(xs: Iterable[Tuple[Key, Value]]) -> MapTree[Key, Value]:
    ie = builtins.iter(xs)
    return mk_from_iterator(empty, ie)


# Imperative left-to-right iterators.


# collapseLHS:
# a) Always returns either [] or a list starting with MapOne.
# b) The "fringe" of the set stack is unchanged.
def collapseLHS(stack: Block[MapTree[Key, Value]]) -> Block[MapTree[Key, Value]]:
    if stack.is_empty():
        return block.empty
    m, rest = stack.head(), stack.tail()
    for m2 in m.to_list():
        if isinstance(m2, MapTreeNode):
            mn = m2
            tree = Some(MapTreeLeaf(mn.key, mn.value))
            return collapseLHS(rest.cons(mn.right).cons(tree).cons(mn.left))
        else:
            return stack
    else:
        return collapseLHS(rest)


class MkIterator(Iterator[Tuple[Key, Value]]):
    def __init__(self, m: MapTree[Key, Value]) -> None:
        self.stack = collapseLHS(block.singleton(m))

    def __next__(self) -> Tuple[Key, Value]:
        if not self.stack:
            raise StopIteration

        rest = self.stack.tail()
        for m in self.stack.head():
            if isinstance(m, MapTreeNode):
                failwith(
                    "Please report error: Map iterator, unexpected stack for next()"
                )
            else:
                self.stack = collapseLHS(rest)
                return m.key, m.value
        else:
            failwith("Please report error: Map iterator, unexpected stack for next()")


def not_started() -> None:
    failwith("enumeration not started")


def already_finished():
    failwith("enumeration already finished")


def mk_iterator(m: MapTree[Key, Value]) -> Iterator[Tuple[Key, Value]]:
    return MkIterator(m)


def to_seq(s: MapTree[Key, Value]) -> Iterable[Tuple[Key, Value]]:
    it = mk_iterator(s)

    def folder(it: Iterator[Tuple[Key, Value]]):
        try:
            current = next(it)
        except StopIteration:
            return Nothing
        else:
            return Some((current, it))

    return pipe(it, seq.unfold(folder))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['expression',
 'expression.collections',
 'expression.core',
 'expression.effect',
 'expression.extra',
 'expression.extra.option',
 'expression.extra.result',
 'expression.system']

package_data = \
{'': ['*']}

install_requires = \
['typing-extensions>=4.1.1,<5.0.0']

setup_kwargs = {
    'name': 'expression',
    'version': '4.2.3',
    'description': 'Practical functional programming for Python 3.9+',
    'long_description': '# Expression\n\n[![PyPI](https://img.shields.io/pypi/v/expression.svg)](https://pypi.python.org/pypi/Expression)\n![Python package](https://github.com/cognitedata/expression/workflows/Python%20package/badge.svg)\n![Upload Python Package](https://github.com/cognitedata/expression/workflows/Upload%20Python%20Package/badge.svg)\n[![Documentation Status](https://readthedocs.org/projects/expression/badge/?version=latest)](https://expression.readthedocs.io/en/latest/?badge=latest)\n[![codecov](https://codecov.io/gh/cognitedata/expression/branch/main/graph/badge.svg)](https://codecov.io/gh/cognitedata/expression)\n\n> Pragmatic functional programming\n\nExpression aims to be a solid, type-safe, pragmatic, and high performance\nlibrary for frictionless and practical functional programming in Python 3.9+.\n\nBy pragmatic, we mean that the goal of the library is to use simple abstractions\nto enable you to do practical and productive functional programming in Python\n(instead of being a [Monad tutorial](https://github.com/dbrattli/OSlash)).\n\nPython is a multi-paradigm programming language that also supports functional\nprogramming constructs such as functions, higher-order functions, lambdas, and\nin many ways favors composition over inheritance.\n\n> Better Python with F#\n\nExpression tries to make a better Python by providing several functional\nfeatures inspired by [F#](https://fsharp.org). This serves several\npurposes:\n\n- Enable functional programming in a Pythonic way, i.e., make sure we are not\n  over-abstracting things. Expression will not require purely functional\n  programming as would a language like Haskell.\n- Everything you learn with Expression can also be used with F#. Learn F# by\n  starting in a programming language they already know. Perhaps get inspired to\n  also [try out F#](https://aka.ms/fsharphome) by itself.\n- Make it easier for F# developers to use Python when needed, and re-use many\n  of the concepts and abstractions they already know and love.\n\nExpression will enable you to work with Python using many of the same\nprogramming concepts and abstractions. This enables concepts such as [Railway\noriented programming](https://fsharpforfunandprofit.com/rop/) (ROP) for better\nand predictable error handling. Pipelining for workflows, computational\nexpressions, etc.\n\n> _Expressions evaluate to a value. Statements do something._\n\nF# is a functional programming language for .NET that is succinct (concise,\nreadable, and type-safe) and kind of\n[Pythonic](https://docs.python.org/3/glossary.html). F# is in many ways very\nsimilar to Python, but F# can also do a lot of things better than Python:\n\n- Strongly typed, if it compiles it usually works making refactoring much\n  safer. You can trust the type-system. With [mypy](http://mypy-lang.org/) or\n  [Pylance](https://github.com/microsoft/pylance-release) you often wonder who\n  is right and who is wrong.\n- Type inference, the compiler deduces types during compilation\n- Expression based language\n\n## Getting Started\n\nYou can install the latest `expression` from PyPI by running `pip` (or\n`pip3`). Note that `expression` only works for Python 3.9+.\n\n```console\n> pip3 install expression\n```\n\n## Goals\n\n- Industrial strength library for functional programming in Python.\n- The resulting code should look and feel like Python\n  ([PEP-8](https://www.python.org/dev/peps/pep-0008/)). We want to make a\n  better Python, not some obscure DSL or academic Monad tutorial.\n- Provide pipelining and pipe friendly methods. Compose all the things!\n- Dot-chaining on objects as an alternative syntax to pipes.\n- Lower the cognitive load on the programmer by:\n  - Avoid currying, not supported in Python by default and not a well known\n    concept by Python programmers.\n  - Avoid operator (`|`, `>>`, etc) overloading, this usually confuses more\n    than it helps.\n  - Avoid recursion. Recursion is not normally used in Python and any use of it\n    should be hidden within the SDK.\n- Provide [type-hints](https://docs.python.org/3/library/typing.html) for all\n  functions and methods.\n- Support PEP 634 and structural pattern matching.\n- Code must pass strict static type checking by\n  [pylance](https://devblogs.microsoft.com/python/announcing-pylance-fast-feature-rich-language-support-for-python-in-visual-studio-code/).\n  Pylance is awesome, use it!\n- [Pydantic](https://pydantic-docs.helpmanual.io/) friendly data types. Use Expression\n  types as part of your Pydantic data model and (de)serialize to/from JSON.\n\n## Supported features\n\nExpression will never provide you with all the features of F# and .NET. We are\nproviding a few of the features we think are useful, and will add more\non-demand as we go along.\n\n- **Pipelining** - for creating workflows.\n- **Composition** - for composing and creating new operators.\n- **Fluent or Functional** syntax, i.e., dot chain or pipeline operators.\n- **Pattern Matching** - an alternative flow control to `if-elif-else`.\n- **Error Handling** - Several error handling types.\n  - **Option** - for optional stuff and better `None` handling.\n  - **Result** - for better error handling and enables railway-oriented\n    programming in Python.\n  - **Try** - a simpler result type that pins the error to an Exception.\n- **Collections** - immutable collections.\n  - **TypedArray** - a generic array type that abstracts the details of\n    `bytearray`, `array.array` and `list` modules.\n  - **Sequence** - a better\n    [itertools](https://docs.python.org/3/library/itertools.html) and\n    fully compatible with Python iterables.\n  - **Block** - a frozen and immutable list type.\n  - **Map** - a frozen and immutable dictionary type.\n  - **AsyncSeq** - Asynchronous iterables.\n  - **AsyncObservable** - Asynchronous observables. Provided separately\n    by [aioreactive](https://github.com/dbrattli/aioreactive).\n- **Data Modelling** - sum and product types\n  - **TaggedUnion** - A tagged (discriminated) union type.\n- **Parser Combinators** - A recursive decent string parser combinator\n  library.\n- **Effects**: - lightweight computational expressions for Python. This\n  is amazing stuff.\n  - **option** - an optional world for working with optional values.\n  - **result** - an error handling world for working with result values.\n- **Mailbox Processor**: for lock free programming using the [Actor\n  model](https://en.wikipedia.org/wiki/Actor_model).\n- **Cancellation Token**: for cancellation of asynchronous (and\n  synchronous) workflows.\n- **Disposable**: For resource management.\n\n### Pipelining\n\nExpression provides a `pipe` function similar to `|>` in F#. We don\'t want to\noverload any Python operators, e.g., `|` so `pipe` is a plain old function taking\nN-arguments, and will let you pipe a value through any number of functions.\n\n```python\nfrom expression import pipe\n\nv = 1\nfn = lambda x: x + 1\ngn = lambda x: x * 2\n\nassert pipe(v, fn, gn) == gn(fn(v))\n```\n\nExpression objects (e.g., `Some`, `Seq`, `Result`) also have a `pipe` method, so you can dot chain pipelines\ndirectly on the object:\n\n```python\nfrom expression import Some\n\nv = Some(1)\nfn = lambda x: x.map(lambda y: y + 1)\ngn = lambda x: x.map(lambda y: y * 2)\n\nassert v.pipe(fn, gn) == gn(fn(v))\n```\n\nSo for example with sequences you may create sequence transforming\npipelines:\n\n```python\nfrom expression.collections import seq, Seq\n\nxs = Seq.of(9, 10, 11)\nys = xs.pipe(\n    seq.map(lambda x: x * 10),\n    seq.filter(lambda x: x > 100),\n    seq.fold(lambda s, x: s + x, 0)\n)\n\nassert ys == 110\n```\n\n### Composition\n\nFunctions may even be composed directly into custom operators:\n\n```python\nfrom expression import compose\nfrom expression.collections import seq, Seq\n\nxs = Seq.of(9, 10, 11)\ncustom = compose(\n    seq.map(lambda x: x * 10),\n    seq.filter(lambda x: x > 100),\n    seq.fold(lambda s, x: s + x, 0)\n)\nys = custom(xs)\n\nassert ys == 110\n```\n\n### Fluent and Functional\n\nExpression can be used both with a fluent or functional syntax (or both.)\n\n#### Fluent syntax\n\nThe fluent syntax uses methods and is very compact. But it might get you into\ntrouble for large pipelines since it\'s not a natural way of adding line breaks.\n\n```python\nfrom expression.collections import Seq\n\nxs = Seq.of(1, 2, 3)\nys = xs.map(lambda x: x * 100).filter(lambda x: x > 100).fold(lambda s, x: s + x, 0)\n```\n\nNote that fluent syntax is probably the better choice if you use mypy\nfor type checking since mypy may have problems inferring types through\nlarger pipelines.\n\n#### Functional syntax\n\nThe functional syntax is a bit more verbose but you can easily add new\noperations on new lines. The functional syntax is great to use together\nwith pylance/pyright.\n\n```python\nfrom expression import pipe\nfrom expression.collections import seq, Seq\n\nxs = Seq.of(1, 2, 3)\nys = pipe(xs,\n    seq.map(lambda x: x * 100),\n    seq.filter(lambda x: x > 100),\n    seq.fold(lambda s, x: s + x, 0),\n)\n```\n\nBoth fluent and functional syntax may be mixed and even pipe can be used\nfluently.\n\n```python\nfrom expression.collections import seq, Seq\nxs = Seq.of(1, 2, 3).pipe(seq.map(...))\n```\n\n### Option\n\nThe `Option` type is used when a function or method cannot produce a meaningful\noutput for a given input.\n\nAn option value may have a value of a given type, i.e., `Some(value)`, or it might\nnot have any meaningful value, i.e., `Nothing`.\n\n```python\nfrom expression import Some, Nothing, Option\n\ndef keep_positive(a: int) -> Option[int]:\n    if a > 0:\n        return Some(a)\n\n    return Nothing\n```\n\n```python\nfrom expression import Option, Ok\ndef exists(x : Option[int]) -> bool:\n    match x:\n        case Some(_):\n            return True\n    return False\n```\n\n### Option as an effect\n\nEffects in Expression is implemented as specially decorated coroutines\n([enhanced generators](https://www.python.org/dev/peps/pep-0342/)) using\n`yield`, `yield from` and `return` to consume or generate optional values:\n\n```python\nfrom expression import effect, Some\n\n@effect.option[int]()\ndef fn():\n    x = yield 42\n    y = yield from Some(43)\n\n    return x + y\n\nxs = fn()\n```\n\nThis enables ["railway oriented\nprogramming"](https://fsharpforfunandprofit.com/rop/), e.g., if one part of the\nfunction yields from `Nothing` then the function is side-tracked\n(short-circuit) and the following statements will never be executed. The end\nresult of the expression will be `Nothing`. Thus results from such an option\ndecorated function can either be `Ok(value)` or `Error(error_value)`.\n\n```python\nfrom expression import effect, Some, Nothing\n\n@effect.option[int]()\ndef fn():\n    x = yield from Nothing # or a function returning Nothing\n\n    # -- The rest of the function will never be executed --\n    y = yield from Some(43)\n\n    return x + y\n\nxs = fn()\nassert xs is Nothing\n```\n\nFor more information about options:\n\n- [Tutorial](https://expression.readthedocs.io/en/latest/tutorial/optional_values.html)\n- [API reference](https://expression.readthedocs.io/en/latest/reference/option.html)\n\n### Result\n\nThe `Result[T, TError]` type lets you write error-tolerant code that can be\ncomposed. A Result works similar to `Option`, but lets you define the value used\nfor errors, e.g., an exception type or similar. This is great when you want to\nknow why some operation failed (not just `Nothing`). This type serves the same\npurpose of an `Either` type where `Left` is used for the error condition and `Right`\nfor a success value.\n\n```python\nfrom expression import effect, Ok, Result\n\n@effect.result[int, Exception]()\ndef fn():\n    x = yield from Ok(42)\n    y = yield from Ok(10)\n    return x + y\n\nxs = fn()\nassert isinstance(xs, Result)\n```\n\nA simplified type called `Try` is also available. It\'s a result type that is\npinned to `Exception` i.e., `Result[TSource, Exception]`.\n\n### Sequence\n\nSequences is a thin wrapper on top of iterables and contains operations for working with\nPython iterables. Iterables are immutable by design, and perfectly suited for functional\nprogramming.\n\n```python\nimport functools\nfrom expression import pipe\nfrom expression.collections import seq\n\n# Normal python way. Nested functions are hard to read since you need to\n# start reading from the end of the expression.\nxs = range(100)\nys = functools.reduce(lambda s, x: s + x, filter(lambda x: x > 100, map(lambda x: x * 10, xs)), 0)\n\n# With Expression, you pipe the result, so it flows from one operator to the next:\nzs = pipe(\n    xs,\n    seq.map(lambda x: x * 10),\n    seq.filter(lambda x: x > 100),\n    seq.fold(lambda s, x: s + x, 0),\n)\nassert ys == zs\n```\n\n## Tagged Unions\n\nTagged Unions (aka discriminated unions) may look similar to normal Python Unions. But\nthey are [different](https://stackoverflow.com/a/61646841) in that the operands in a\ntype union `(A | B)` are both types, while the cases in a tagged union type `U = A | B`\nare both constructors for the type U and are not types themselves. One consequence is\nthat tagged unions can be nested in a way union types might not.\n\nIn Expression you make a tagged union by defining your type as a sub-class of\n`TaggedUnion` with the appropriate generic types that this union represent for\neach case. Then you define static or class-method constructors for creating each of the\ntagged union cases.\n\n```python\nfrom dataclasses import dataclass\nfrom expression import TaggedUnion, tag\n\n@dataclass\nclass Rectangle:\n    width: float\n    length: float\n\n@dataclass\nclass Circle:\n    radius: float\n\nclass Shape(TaggedUnion):\n    RECTANGLE = tag(Rectangle)\n    CIRCLE = tag(Circle)\n\n    @staticmethod\n    def rectangle(width: float, length: float) -> Shape:\n        return Shape(Shape.RECTANGLE, Rectangle(width, length))\n\n    @staticmethod\n    def circle(radius: float) -> Shape:\n        return Shape(Shape.CIRCLE, Circle(radius))\n```\n\nNow you may pattern match the shape to get back the actual value:\n\n```python\n    from expression import match\n\n    shape = Shape.Rectangle(2.3, 3.3)\n\n    match shape:\n        case Shape(value=Rectangle(width=2.3)):\n            assert shape.value.width == 2.3\n        case _:\n            assert False\n```\n\n## Notable differences between Expression and F#\n\nIn F# modules are capitalized, in Python they are lowercase\n([PEP-8](https://www.python.org/dev/peps/pep-0008/#package-and-module-names)).\nE.g in F# `Option` is both a module (`OptionModule` internally) and a\ntype. In Python the module is `option` and the type is capitalized i.e\n`Option`.\n\nThus in Expression you use `option` as the module to access module functions\nsuch as `option.map` and the name `Option` for the type itself.\n\n```pycon\n>>> from expression import Option, option\n>>> Option\n<class \'expression.core.option.Option\'>\n>>> option\n<module \'expression.core.option\' from \'/Users/dbrattli/Developer/Github/Expression/expression/core/option.py\'>\n```\n\n## Common Gotchas and Pitfalls\n\nA list of common problems and how you may solve it:\n\n### Expression is missing the function/operator I need\n\nRemember that everything is just a function, so you can easily implement\na custom function yourself and use it with Expression. If you think the\nfunction is also usable for others, then please open a PR to include it\nwith Expression.\n\n## Resources and References\n\nA collection of resources that were used as reference and inspiration\nfor creating this library.\n\n- F# (http://fsharp.org)\n- Get Started with F# (https://aka.ms/fsharphome)\n- F# as a Better Python - Phillip Carter - NDC Oslo 2020\n  (https://www.youtube.com/watch?v=_QnbV6CAWXc)\n- OSlash (https://github.com/dbrattli/OSlash)\n- RxPY (https://github.com/ReactiveX/RxPY)\n- PEP 8 -- Style Guide for Python Code (https://www.python.org/dev/peps/pep-0008/)\n- PEP 342 -- Coroutines via Enhanced Generators\n  (https://www.python.org/dev/peps/pep-0342/)\n- PEP 380 -- Syntax for Delegating to a Subgenerator\n  (https://www.python.org/dev/peps/pep-0380)\n- PEP 479 -- Change StopIteration handling inside generators (https://www.python.org/dev/peps/pep-0479/)\n- PEP 634 -- Structural Pattern Matching (https://www.python.org/dev/peps/pep-0634/)\n- Thunks, Trampolines and Continuation Passing\n  (https://jtauber.com/blog/2008/03/30/thunks,_trampolines_and_continuation_passing/)\n- Tail Recursion Elimination\n  (http://neopythonic.blogspot.com/2009/04/tail-recursion-elimination.html)\n- Final Words on Tail Calls\n  (http://neopythonic.blogspot.com/2009/04/final-words-on-tail-calls.html)\n- Python is the Haskell You Never Knew You Had: Tail Call Optimization\n  (https://sagnibak.github.io/blog/python-is-haskell-tail-recursion/)\n\n## How-to Contribute\n\nYou are very welcome to contribute with suggestions or PRs :heart_eyes: It is\nnice if you can try to align the code and naming with F# modules, functions,\nand documentation if possible. But submit a PR even if you should feel unsure.\n\nCode, doc-strings, and comments should also follow the [Google Python Style\nGuide](https://google.github.io/styleguide/pyguide.html).\n\nCode checks are done using\n\n- [Black](https://github.com/psf/black)\n- [flake8](https://github.com/PyCQA/flake8)\n- [isort](https://github.com/PyCQA/isort)\n\nTo run code checks on changed files every time you commit, install the pre-commit hooks\nby running:\n\n```console\n> pre-commit install\n```\n\n## Code of Conduct\n\nThis project follows https://www.contributor-covenant.org, see our [Code\nof\nConduct](https://github.com/cognitedata/Expression/blob/main/CODE_OF_CONDUCT.md).\n\n## License\n\nMIT, see [LICENSE](https://github.com/cognitedata/Expression/blob/main/LICENSE).\n',
    'author': 'Dag Brattli',
    'author_email': 'dag.brattli@cognite.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/cognitedata/Expression',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4',
}


setup(**setup_kwargs)



"""
Pipe module.

Contains pipe function including necessary overloads to get the type-hints right.

Example:
    >>> from expression import pipe
    >>>
    >>> v = 1
    >>> fn = lambda x: x + 1
    >>> gn = lambda x: x * 2
    >>>
    >>> assert pipe(v, fn, gn) == gn(fn(v))
"""
from typing import Any, Callable, Tuple, TypeVar, overload

from .compose import compose
from .misc import starid

_A = TypeVar("_A")
_B = TypeVar("_B")
_C = TypeVar("_C")
_D = TypeVar("_D")
_E = TypeVar("_E")
_F = TypeVar("_F")
_G = TypeVar("_G")
_H = TypeVar("_H")
_T = TypeVar("_T")
_J = TypeVar("_J")


@overload
def pipe(__value: _A) -> _A:
    ...


@overload
def pipe(__value: _A, __fn1: Callable[[_A], _B]) -> _B:
    ...


@overload
def pipe(__value: _A, __fn1: Callable[[_A], _B], __fn2: Callable[[_B], _C]) -> _C:
    ...


@overload
def pipe(
    __value: _A,
    __fn1: Callable[[_A], _B],
    __fn2: Callable[[_B], _C],
    __fn3: Callable[[_C], _D],
) -> _D:
    ...


@overload
def pipe(
    __value: _A,
    __fn1: Callable[[_A], _B],
    __fn2: Callable[[_B], _C],
    __fn3: Callable[[_C], _D],
    __fn4: Callable[[_D], _E],
) -> _E:
    ...


@overload
def pipe(
    __value: _A,
    __fn1: Callable[[_A], _B],
    __fn2: Callable[[_B], _C],
    __fn3: Callable[[_C], _D],
    __fn4: Callable[[_D], _E],
    __fn5: Callable[[_E], _F],
) -> _F:
    ...


@overload
def pipe(
    __value: _A,
    __fn1: Callable[[_A], _B],
    __fn2: Callable[[_B], _C],
    __fn3: Callable[[_C], _D],
    __fn4: Callable[[_D], _E],
    __fn5: Callable[[_E], _F],
    __fn6: Callable[[_F], _G],
) -> _G:
    ...


@overload
def pipe(
    __value: _A,
    __fn1: Callable[[_A], _B],
    __fn2: Callable[[_B], _C],
    __fn3: Callable[[_C], _D],
    __fn4: Callable[[_D], _E],
    __fn5: Callable[[_E], _F],
    __fn6: Callable[[_F], _G],
    __fn7: Callable[[_G], _H],
) -> _H:
    ...


@overload
def pipe(
    __value: _A,
    __fn1: Callable[[_A], _B],
    __fn2: Callable[[_B], _C],
    __fn3: Callable[[_C], _D],
    __fn4: Callable[[_D], _E],
    __fn5: Callable[[_E], _F],
    __fn6: Callable[[_F], _G],
    __fn7: Callable[[_G], _H],
    __fn8: Callable[[_H], _T],
) -> _T:
    ...


@overload
def pipe(
    __value: _A,
    __fn1: Callable[[_A], _B],
    __fn2: Callable[[_B], _C],
    __fn3: Callable[[_C], _D],
    __fn4: Callable[[_D], _E],
    __fn5: Callable[[_E], _F],
    __fn6: Callable[[_F], _G],
    __fn7: Callable[[_G], _H],
    __fn8: Callable[[_H], _T],
    __fn9: Callable[[_T], _J],
) -> _J:
    ...


def pipe(__value: Any, *fns: Callable[[Any], Any]) -> Any:
    """Functional pipe (`|>`)

    Allows the use of function argument on the left side of the
    function.

    Example:
        >>> pipe(x, fn) == __fn(x)  # Same as x |> fn
        >>> pipe(x, fn, gn) == gn(fn(x))  # Same as x |> fn |> gn
        ...
    """

    return compose(*fns)(__value)


@overload
def pipe2(__values: Tuple[_A, _B]) -> Tuple[_A, _B]:
    ...


@overload
def pipe2(__values: Tuple[_A, _B], __fn1: Callable[[_A], Callable[[_B], _C]]) -> _C:
    ...


@overload
def pipe2(
    __values: Tuple[_A, _B],
    __fn1: Callable[[_A], Callable[[_B], _C]],
    __fn2: Callable[[_C], _D],
) -> _D:
    ...


@overload
def pipe2(
    __values: Tuple[_A, _B],
    __fn1: Callable[[_A], Callable[[_B], _C]],
    __fn2: Callable[[_C], _D],
    __fn3: Callable[[_D], _E],
) -> _E:
    ...


def pipe2(__values: Any, *fns: Any) -> Any:
    return pipe(fns[0](__values[0])(__values[1]), *fns[1:]) if fns else __values


def pipe3(__values: Any, *fns: Any) -> Any:
    return (
        pipe(fns[0](__values[0])(__values[1])(__values[2]), *fns[1:])
        if fns
        else __values
    )


def starpipe(args: Any, *fns: Callable[..., Any]):
    """Functional pipe_n (`||>`, `||>`, `|||>`, etc)

    Allows the use of function arguments on the left side of the
    function. Calls the function with tuple arguments unpacked.

    Example:
        >>> starpipe((x, y), __fn) == __fn(x, y)  # Same as (x, y) ||> __fn
        >>> starpipe((x, y), __fn, gn) == gn(fn(x))  # Same as (x, y) ||> __fn |> gn
        ...
    """

    fn = fns[0] if len(fns) else starid

    return compose(*fns[1:])(fn(*args))


class PipeMixin:
    """A pipe mixin class that enabled a class to use pipe fluently."""

    @overload
    def pipe(self: _A, __fn1: Callable[[_A], _B]) -> _B:
        ...

    @overload
    def pipe(self: _A, __fn1: Callable[[_A], _B], __fn2: Callable[[_B], _C]) -> _C:
        ...

    @overload
    def pipe(
        self: _A,
        __fn1: Callable[[_A], _B],
        __fn2: Callable[[_B], _C],
        __fn3: Callable[[_C], _D],
    ) -> _D:
        ...

    @overload
    def pipe(
        self: _A,
        __fn1: Callable[[_A], _B],
        __fn2: Callable[[_B], _C],
        __fn3: Callable[[_C], _D],
        __fn4: Callable[[_D], _E],
    ) -> _E:
        ...

    @overload
    def pipe(
        self: _A,
        __fn1: Callable[[_A], _B],
        __fn2: Callable[[_B], _C],
        __fn3: Callable[[_C], _D],
        __fn4: Callable[[_D], _E],
        __fn5: Callable[[_E], _F],
    ) -> _F:
        ...

    @overload
    def pipe(
        self: _A,
        __fn1: Callable[[_A], _B],
        __fn2: Callable[[_B], _C],
        __fn3: Callable[[_C], _D],
        __fn4: Callable[[_D], _E],
        __fn5: Callable[[_E], _F],
        __fn6: Callable[[_F], _G],
    ) -> _G:
        ...

    def pipe(self, *args: Any) -> Any:
        """Pipe the left side object through the given functions."""
        return pipe(self, *args)


__all__ = ["pipe", "pipe2", "pipe3", "PipeMixin", "starpipe"]



from typing import Any, Callable, Iterable, TypeVar

from expression.collections import seq
from expression.core import Builder

_TSource = TypeVar("_TSource")
_TResult = TypeVar("_TResult")


class SeqBuilder(Builder[_TSource, Iterable[Any]]):
    def bind(
        self, xs: Iterable[_TSource], fn: Callable[[_TSource], Iterable[_TResult]]
    ) -> Iterable[_TResult]:
        for x in xs:
            return fn(x)
        return []

    def return_(self, x: _TSource) -> Iterable[_TSource]:
        return seq.singleton(x)

    def return_from(self, xs: Iterable[_TSource]) -> Iterable[_TSource]:
        return xs

    def combine(
        self, xs: Iterable[_TSource], ys: Iterable[_TSource]
    ) -> Iterable[_TSource]:
        ret = seq.concat(xs, ys)
        return ret

    def zero(self) -> Iterable[_TSource]:
        return seq.empty


__all__ = ["SeqBuilder"]



"""A typed array module.

This module provides an typed array type `TypedArray` and a set of
useful methods and functions for working with the array. The internal
backing storage is `bytearray`, `array.array` or `list` depending on
type of input.

"""
from __future__ import annotations

import array
import builtins
import functools
from enum import Enum
from typing import (
    Any,
    Callable,
    Iterable,
    Iterator,
    List,
    MutableSequence,
    Optional,
    Tuple,
    TypeVar,
    cast,
)

from expression.core import (
    Nothing,
    Option,
    PipeMixin,
    Some,
    SupportsLessThan,
    SupportsSum,
    curry_flip,
    pipe,
)

from . import seq

_TSource = TypeVar("_TSource")
_TResult = TypeVar("_TResult")
_TState = TypeVar("_TState")
_TSourceSortable = TypeVar("_TSourceSortable", bound=SupportsLessThan)

_TSourceSum = TypeVar("_TSourceSum", bound=SupportsSum)
_Array = List[_TSource] | MutableSequence[_TSource]


class int8(int):
    ...


class int16(int):
    ...


class int32(int):
    ...


class int64(int):
    ...


class uint8(int):
    ...


class uint16(int):
    ...


class uint32(int):
    ...


class uint64(int):
    ...


class float32(float):
    ...


class float64(float):
    ...


class double(float):
    ...


class TypeCode(Enum):
    Byte = "y"
    Int8 = "b"
    Int16 = "h"
    Int32 = "l"
    Int64 = "q"
    Uint8 = "B"
    Uint16 = "H"
    Uint32 = "L"
    Uint64 = "Q"
    Double = "d"
    Float = "f"
    # String = "s"
    Any = "a"


def array_from_typecode(
    type_code: TypeCode, initializer: Optional[Iterable[Any]]
) -> _Array[Any]:
    arr: _Array[Any] = list(initializer if initializer else [])
    if type_code == TypeCode.Byte:
        return bytearray(arr)
    elif type_code == TypeCode.Int8:
        return array.array("b", arr)
    elif type_code == TypeCode.Uint8:
        return array.array("B", arr)
    elif type_code == TypeCode.Int16:
        return array.array("h", arr)
    elif type_code == TypeCode.Uint16:
        return array.array("H", arr)
    elif type_code == TypeCode.Int32:
        return array.array("l", arr)
    elif type_code == TypeCode.Uint32:
        return array.array("L", arr)
    elif type_code == TypeCode.Int64:
        return array.array("q", arr)
    elif type_code == TypeCode.Uint64:
        return array.array("Q", arr)
    elif type_code == TypeCode.Float:
        return array.array("f", arr)
    elif type_code == TypeCode.Double:
        return array.array("d", arr)
    return arr


def array_from_initializer(
    initializer: Optional[Iterable[Any]] = None,
) -> Tuple[_Array[Any], TypeCode]:
    # Use list as the default array
    arr: _Array[Any] = list(initializer if initializer else [])
    type_code = TypeCode.Any

    if isinstance(initializer, bytearray):
        arr = initializer
        type_code = TypeCode.Byte
    elif isinstance(initializer, bytes):
        arr = bytearray(initializer)
        type_code = TypeCode.Byte
    elif arr:
        arr0 = arr[0]
        if isinstance(arr0, int8):
            arr = array.array("b", arr)
            type_code = TypeCode.Int8
        elif isinstance(arr0, int16):
            arr = array.array("h", arr)
            type_code = TypeCode.Int16
        elif isinstance(arr0, int32):
            arr = array.array("l", arr)
            type_code = TypeCode.Int32
        elif isinstance(arr0, int64):
            arr = array.array("l", arr)
            type_code = TypeCode.Int64
        elif isinstance(arr0, uint8):
            arr = array.array("B", arr)
            type_code = TypeCode.Uint8
        elif isinstance(arr0, uint16):
            arr = array.array("H", arr)
            type_code = TypeCode.Uint16
        elif isinstance(arr0, uint32):
            arr = array.array("L", arr)
            type_code = TypeCode.Uint32
        elif isinstance(arr0, uint64):
            arr = array.array("Q", arr)
            type_code = TypeCode.Uint64
        elif isinstance(arr0, float32):
            arr = array.array("f", arr)
            type_code = TypeCode.Float
        elif isinstance(arr0, (float64, double)):
            arr = array.array("d", arr)
            type_code = TypeCode.Double

    return arr, type_code


class TypedArray(MutableSequence[_TSource], PipeMixin):
    __match_args__ = (
        "initializer",
        "typecode",
    )

    def __init__(
        self,
        initializer: Optional[Iterable[_TSource]] = None,
        typecode: Optional[TypeCode] = None,
    ) -> None:
        if typecode:
            arr = array_from_typecode(typecode, initializer)
        else:
            arr, typecode = array_from_initializer(initializer)

        self.value: _Array[Any] = arr
        self.typecode = typecode

    def map(self, mapping: Callable[[_TSource], _TResult]) -> TypedArray[_TResult]:
        result = builtins.map(mapping, self.value)
        return TypedArray(result)

    def choose(
        self, chooser: Callable[[_TSource], Option[_TResult]]
    ) -> TypedArray[_TResult]:
        """Choose items from the list.

        Applies the given function to each element of the list. Returns
        the list comprised of the results x for each element where the
        function returns `Some(x)`.

        Args:
            chooser: The function to generate options from the elements.

        Returns:
            The list comprising the values selected from the chooser
            function.
        """

        def mapper(x: _TSource) -> TypedArray[_TResult]:
            return TypedArray(chooser(x).to_seq())

        return self.collect(mapper)

    def collect(
        self, mapping: Callable[[_TSource], TypedArray[_TResult]]
    ) -> TypedArray[_TResult]:
        mapped = builtins.map(mapping, self.value)
        xs = (y for x in mapped for y in x)
        return TypedArray(xs)

    def insert(self, index: int, value: _TSource) -> None:
        self.value.insert(index, value)

    def is_empty(self) -> bool:
        """Return `True` if list is empty."""

        return not self.value

    @property
    @classmethod
    def empty(cls) -> TypedArray[Any]:
        """Returns empty array."""

        return TypedArray()

    def filter(self, predicate: Callable[[_TSource], bool]) -> TypedArray[_TSource]:
        """Filter list.

        Returns a new collection containing only the elements of the
        collection for which the given predicate returns `True`.

        Args:
            predicate: The function to test the input elements.

        Returns:
            A list containing only the elements that satisfy the
            predicate.
        """
        return TypedArray(
            builtins.filter(predicate, self.value), typecode=self.typecode
        )

    def fold(
        self, folder: Callable[[_TState, _TSource], _TState], state: _TState
    ) -> _TState:
        """Applies a function to each element of the array,
        threading an accumulator argument through the computation. Take
        the second argument, and apply the function to it and the first
        element of the list. Then feed this result into the function
        along with the second element and so on. Return the final
        result. If the input function is f and the elements are i0...iN
        then computes f (... (f s i0) i1 ...) iN.

        Args:
            folder: The function to update the state given the input
                elements.

            state: The initial state.

        Returns:
            Partially applied fold function that takes the source list
            and returns the final state value.
        """
        return functools.reduce(folder, self, state)

    def forall(self, predicate: Callable[[_TSource], bool]) -> bool:
        """Tests if all elements of the collection satisfy the given
        predicate.

        Args:
            predicate: The function to test the input elements.

        Returns:
            True if all of the elements satisfy the predicate.
        """
        return all(predicate(x) for x in self)

    def head(self) -> _TSource:
        """Returns the first element of the list.

        Args:
            source: The input list.

        Returns:
            The first element of the list.

        Raises:
            ValueError: Thrown when the list is empty.
        """

        head, *_ = self
        return head

    def indexed(self, start: int = 0) -> TypedArray[Tuple[int, _TSource]]:
        """Returns a new array whose elements are the corresponding
        elements of the input array paired with the index (from `start`)
        of each element.

        Args:
            start: Optional index to start from. Defaults to 0.

        Returns:
            The list of indexed elements.
        """
        return of_seq(enumerate(self))

    def item(self, index: int) -> _TSource:
        """Indexes into the list. The first element has index 0.

        Args:
            index: The index to retrieve.

        Returns:
            The value at the given index.
        """
        return self.value[index]

    @staticmethod
    def of(*args: _TSource) -> TypedArray[_TSource]:
        """Create list from a number of arguments."""
        return TypedArray((*args,))

    @staticmethod
    def of_seq(xs: Iterable[_TSource]) -> TypedArray[_TSource]:
        """Create list from iterable sequence."""
        return TypedArray((*xs,))

    def skip(self, count: int) -> TypedArray[_TSource]:
        """Returns the array after removing the first N elements.

        Args:
            count: The number of elements to skip.

        Returns:
            The array after removing the first N elements.
        """
        return TypedArray(self.value[count:])

    def skip_last(self, count: int) -> TypedArray[_TSource]:
        return TypedArray(self.value[:-count])

    def sort(
        self: TypedArray[_TSourceSortable], reverse: bool = False
    ) -> TypedArray[_TSourceSortable]:
        """Sort array directly.

        Returns a new sorted collection.

        Args:
            reverse: Sort in reversed order.

        Returns:
            A sorted array.
        """
        return TypedArray(builtins.sorted(self.value, reverse=reverse))

    def sort_with(
        self, func: Callable[[_TSource], Any], reverse: bool = False
    ) -> TypedArray[_TSource]:
        """Sort array with supplied function.

        Returns a new sorted collection.

        Args:
            func: The function to extract a comparison key from each element in list.
            reverse: Sort in reversed order.

        Returns:
            A sorted array.
        """
        return TypedArray(builtins.sorted(self.value, key=func, reverse=reverse))

    def sum(self: TypedArray[_TSourceSum]) -> int:
        return sum(self)

    def sum_by(self, projection: Callable[[_TSource], _TSourceSum]) -> int:
        return pipe(self, sum_by(projection))

    def tail(self) -> TypedArray[_TSource]:
        """Return tail of List."""

        _, *tail = self.value
        return TypedArray(tail)

    def take(self, count: int) -> TypedArray[_TSource]:
        """Returns the first N elements of the list.

        Args:
            count: The number of items to take.

        Returns:
            The result list.
        """
        return TypedArray(self.value[:count])

    def take_last(self, count: int) -> TypedArray[_TSource]:
        """Returns a specified number of contiguous elements from the
        end of the list.

        Args:
            count: The number of items to take.

        Returns:
            The result list.
        """
        return TypedArray(self.value[-count:])

    def try_head(self) -> Option[_TSource]:
        """Returns the first element of the list, or None if the list is
        empty.
        """
        if self.value:
            value = cast("_TSource", self.value[0])
            return Some(value)

        return Nothing

    @staticmethod
    def unfold(
        generator: Callable[[_TState], Option[Tuple[_TSource, _TState]]], state: _TState
    ) -> TypedArray[_TSource]:
        """Returns a list that contains the elements generated by the
        given computation. The given initial state argument is passed to
        the element generator.

        Args:
            generator: A function that takes in the current state and
                returns an option tuple of the next element of the list
                and the next state value.
            state: The initial state.

        Returns:
            The result list.
        """

        return pipe(state, unfold(generator))

    def __eq__(self, o: Any) -> bool:
        if len(self) != len(o):
            return False

        for i, x in enumerate(self):
            if x != o[i]:
                return False
        return True

    def __len__(self) -> int:
        return len(self.value)

    def __iter__(self) -> Iterator[_TSource]:
        return iter(self.value)

    def __setitem__(self, key: int, value: _TSource) -> None:
        self.value[key] = value

    def __getitem__(self, key: Any) -> Any:
        return self.value[key]

    def __delitem__(self, key: int) -> None:
        del self.value[key]

    def __str__(self) -> str:
        return f"[|{', '.join(self.map(str))}|]"

    def __repr__(self) -> str:
        return str(self)


@curry_flip(1)
def map(
    source: TypedArray[_TSource], mapper: Callable[[_TSource], _TResult]
) -> TypedArray[_TResult]:
    """Map array.

    Builds a new array whose elements are the results of applying
    the given function to each of the elements of the array.

    Args:
        mapper: The function to transform elements from the input array.

    Returns:
        A new array of transformed elements.
    """

    return source.map(mapper)


def empty() -> TypedArray[Any]:
    return TypedArray()


@curry_flip(1)
def filter(
    source: TypedArray[_TSource], predicate: Callable[[_TSource], bool]
) -> TypedArray[_TSource]:
    """Returns a new array containing only the elements of the
    array for which the given predicate returns `True`

    Args:
        predicate: The function to test the input elements.

    Returns:
        Partially applied filter function.
    """

    return source.filter(predicate)


@curry_flip(1)
def fold(
    source: TypedArray[_TSource],
    folder: Callable[[_TState, _TSource], _TState],
    state: _TState,
) -> _TState:
    """Applies a function to each element of the collection, threading
    an accumulator argument through the computation. Take the second
    argument, and apply the function to it and the first element of the
    list. Then feed this result into the function along with the second
    element and so on. Return the final result. If the input function is
    f and the elements are i0...iN then computes f (... (f s i0) i1 ...)
    iN.

    Args:
        folder: The function to update the state given the input
            elements.

        state: The initial state.

    Returns:
        Partially applied fold function that takes the source list
        and returns the final state value.
    """

    return source.fold(folder, state)


@curry_flip(1)
def indexed(
    source: TypedArray[_TSource],
    start: int = 0,
) -> TypedArray[Tuple[int, _TSource]]:
    return source.indexed()


def is_empty(source: TypedArray[Any]) -> bool:
    """Returns `True` if the list is empty, `False` otherwise."""
    return source.is_empty()


@curry_flip(1)
def item(source: TypedArray[_TSource], index: int) -> _TSource:
    return source.item(index)


def of(*args: _TSource) -> TypedArray[_TSource]:
    """Create list from a number of arguments."""
    return TypedArray((*args,))


def of_seq(xs: Iterable[_TSource]) -> TypedArray[_TSource]:
    """Create list from iterable sequence."""
    return TypedArray((*xs,))


def of_option(option: Option[_TSource]) -> TypedArray[_TSource]:
    if isinstance(option, Some):
        return singleton(option.value)
    return empty()


def singleton(value: _TSource) -> TypedArray[_TSource]:
    return TypedArray((value,))


def sum(source: TypedArray[_TSourceSum]) -> int:
    return builtins.sum(source.value)


@curry_flip(1)
def sum_by(
    source: TypedArray[_TSource], projection: Callable[[_TSource], _TSourceSum]
) -> int:
    return builtins.sum(source.map(projection).value)


@curry_flip(1)
def take(source: TypedArray[_TSource], count: int) -> TypedArray[_TSource]:
    """Returns the first N elements of the array.

    Args:
        count: The number of items to take.

    Returns:
        The result array.
    """

    return source.take(count)


@curry_flip(1)
def take_last(source: TypedArray[_TSource], count: int) -> TypedArray[_TSource]:
    """Returns a specified number of contiguous elements from the end of
    the list.

    Args:
        count: The number of items to take.

    Returns:
        The result list.
    """

    return source.take_last(count)


def try_head(source: TypedArray[_TSource]) -> Option[_TSource]:
    """Try to get the first element from the list.

    Returns the first element of the list, or None if the list is empty.

    Args:
        source: The input list.

    Returns:
        The first element of the list or `Nothing`.
    """
    return source.try_head()


@curry_flip(1)
def unfold(
    state: _TState, generator: Callable[[_TState], Option[Tuple[_TSource, _TState]]]
) -> TypedArray[_TSource]:
    """Returns a list that contains the elements generated by the
    given computation. The given initial state argument is passed to
    the element generator.

    Args:
        generator: A function that takes in the current state and
            returns an option tuple of the next element of the list
            and the next state value.
        state: The initial state.

    Returns:
        The result list.
    """

    xs = pipe(state, seq.unfold(generator))
    return TypedArray(xs)


__all__ = [
    "empty",
    "filter",
    "fold",
    "indexed",
    "is_empty",
    "item",
    "map",
    "of_option",
    "of_seq",
    "of",
    "sum",
    "sum_by",
    "singleton",
    "take",
    "take_last",
    "try_head",
    "TypedArray",
    "unfold",
]
